import os
import matplotlib.pyplot as plt
import numpy as np

# PSSPY-related packages
import psspy
import dyntools
import redirect

# Default variables for PSSPY
_i = psspy.getdefaultint()
_f = psspy.getdefaultreal()
_s = psspy.getdefaultchar()

class ProgrammableGenerationUnit(object):
    """Creates an object which monitors frequency, ROCOF and voltage to alter the load at a given bus in response to the
     local bus quantities, emulating increased generation by reducing load consumption."""
    def __init__(self, bus_number, monitor_type = "BSFREQ", d_write = 5, limits = [49.9, 49.8, 49.7], activation_at_limits = [20, 20, 20],
                 time_to_activate = [5.0, 5.0, 5.0], max_time_available = [5.0, 5.0, 5.0], max_output = 10000,
                 response_remains_active = False):
        # Variables required upon initialization
        self.bus_number = bus_number  # Bus at which the PGU is placed
        # self.load_id = load_id  # Don't have to use load_id, this line can be removed?
        self.max_change = max_output  # Total max amount of power this unit can supply
        self.d_write = d_write  # The amount of time-steps between each stop to manually check values and initiate control

        # Storing the present load at the bus when PGU is initialized
        numbers = np.array(psspy.alodbusint(-1, 1, "NUMBER")[1][0])
        self.load_bus_index = np.where(numbers == bus_number)[0]
        self.load_at_start = np.array(psspy.alodbusreal(-1, 1, "MVANOM")[1][0])[self.load_bus_index][0]  # Add a variable for storing the current load at the desired bus so it is remembered between load adjustments

        loads_at_buses = np.array(psspy.aloadint(-1, 1, "NUMBER")[1][0])
        self.nr_of_loads_at_bus = sum(np.where(loads_at_buses == bus_number)[0] > 0)

        self.limits = limits  # The limits at which the unit supplies power, for example freq=49.5 Hz will supply 20 MW
        self.activation_at_limits = activation_at_limits  # How much power the unit supplies at the different sensed limits exceeded
        self.time_to_activate = time_to_activate  # Used for resetting the counter in time_to_activate_counter
        self.time_to_activate_counter = time_to_activate[:]  # Time-delay for each generation step to activate upon sensing
        #  a limit exceeded
        self.max_time_available = max_time_available  # The amount of time the different load/supply steps can be active
        self.active_since_last = np.zeros((len(self.limits), 1), dtype=bool)  # Bools if the different limits are exceeded between each call of the update_pgu() function
        self.response_remains_active = response_remains_active  # True if increased frequency should cause previous
        # low-frequency response to deactivate

    # Variables supplied later as the object is called
        self.time = []  # The time for each measurement
        self.single_channel_data = []  # The raw-data input from the calling function

    # Variables calculated as function is called over and over
        self.current_power_output = 0
        self.channel_time = []
        self.output_power_over_time = []

    # Interaction with UFLS and UVLS?
        self.power_previous_pgu_call = 0  # Load power at the end of last iteration with PGU
        #self.power_current_pgu_call = 0  # Load power at new run with PGU
        #load_now = 0

    # Simulation time
        self.current_time = 0  # The current simulation time when the function is called
        self.time_since_last_call = 0  # The simulation time passed since the last the load was called


        for i in range(len(self.limits)):
            # The values passed from calling function are p.u. frequency deviation from nominal, so limits must be
            # converted to the same format.
            self.limits[i] = (self.limits[i] - 50) / 50

        if self._verify_equal_lengths():
            pass  # No error in initialization
        else:
            self.error  # Some kind of faulty initialization, throw error?

    def update_pgu(self, channel_time, single_channel_data):
        """
            Function called every stopped time-step in PssPy to control for the loads' activation parameters and to
            thereafter step loads if conditions are met.
        :return:
        """
        self.time_since_last_call = channel_time[len(channel_time)-1] - self.current_time
        self.current_time = channel_time[len(channel_time)-1]

        self.channel_time = channel_time  # Superfluous?
        self.single_channel_data = single_channel_data  # Superfluous?


        # Count the amount of remaining time the unit has to supply MWs
        for i in range(len(self.active_since_last)):
            if self.active_since_last[i]:
                self.max_time_available[i] -= self.time_since_last_call

        # Has been constant power output since last function call; save this power output so it can be plotted later
        for i in range(len(channel_time) - len(self.output_power_over_time)):  # For use in plotting power output over time
            self.output_power_over_time.append(self.current_power_output)

        # Check if conditions are exceeded and activate generation as programmed
        self._activate_freq_response()  # Check if bus frequency is exceeded, and at which level

    def plot(self, save_plots_local_folder):
        for i in range(len(self.single_channel_data)):
            self.single_channel_data[i] = self.single_channel_data[i] * 50 + 50  # Convert p.u. freq. deviation to freq

        fig, ax1 = plt.subplots()
        fig.suptitle("PGU "+str(self.bus_number), fontsize=12)
        ax1.plot(self.channel_time, self.single_channel_data, "b-")
        ax1.set_xlabel("Time [s]")
        ax1.set_ylabel("Frequency [Hz]", color="b")
        ax1.tick_params("y", colors="b")
        ax1.grid()

        ax2 = ax1.twinx()
        ax2.plot(self.channel_time, self.output_power_over_time, color="r")
        ax2.set_ylabel("Power output [MW]", color="r")
        ax2.tick_params("y", colors="r")
        # For when I want frequency on left y-axis, MW output on right y-axis
        # https://matplotlib.org/examples/api/two_scales.html
        pass

    def return_bus_number(self):
        return self.bus_number

# Private functions
    def _verify_equal_lengths(self):
        # Verifies that the different input matrices are of equal length
        len1, len2, len3, len4 = len(self.limits), len(self.activation_at_limits), len(self.time_to_activate), len(self.max_time_available)
        return ((len1 == len2) and (len3 == len4) and (len2 == len3))

    def _activate_freq_response(self):
        bus_freq_data = self.single_channel_data

        # Only works for LOW frequencies, not high frequencies
        end_index = len(bus_freq_data)  # The index of the latest input data
        last_freq = bus_freq_data[end_index-1]  # -1 due to zero-indexing
        test_freq = last_freq*50 + 50

        # Count the amount of remaining time till unit activates
        for i in range(len(self.time_to_activate)):
            #if min_freq < self.limits[i]:  # Count down the time until the load should be activated
            if last_freq < self.limits[i]:
                self.time_to_activate_counter[i] -= self.time_since_last_call
            else:  # Limits not exceeded, generation should not be activated and counter is thus reset
                self.time_to_activate_counter[i] = self.time_to_activate[i]

        current_power_output = 0
        for i in range(len(self.activation_at_limits)):
            if self.max_time_available[i] > 0 and self.time_to_activate_counter[i] <= 0:
                current_power_output += self.activation_at_limits[i]  # Activate this power
                self.active_since_last[i] = 1  # Will now be counter on next iteration
            else:
                self.active_since_last[i] = 0
        self.current_power_output = min(current_power_output, self.max_change)  # Hinders output exceeding maximum
        if self.response_remains_active:
            self.current_power_output = max(self.current_power_output, self.power_previous_pgu_call)
        power_adjustment = self.current_power_output - self.power_previous_pgu_call
        if power_adjustment == 0:  # If no power change happens since last iteration: do nothing
            return
        self.power_previous_pgu_call = self.current_power_output
        load_now = np.array(psspy.alodbuscplx(-1, 1, "MVANOM")[1][0])[self.load_bus_index][0].real
        new_load_setpoint = (load_now - power_adjustment)/self.nr_of_loads_at_bus
        for i in range(self.nr_of_loads_at_bus):
            psspy.load_chng_4(self.bus_number, str(i+1), [_i, _i, _i, _i, _i, _i],
                              [new_load_setpoint, _f, _f, _f, _f, _f])


