# Standard Python-packages
import os
import matplotlib.pyplot as plt
import numpy as np

# PSSPY-related packages
import psspy
import dyntools
import redirect

# My own packages
import programmable_generation as pg

# Default variables for PSSPY
_i = psspy.getdefaultint()
_f = psspy.getdefaultreal()
_s = psspy.getdefaultchar()


class PsspyCase(object):
    """Creates an object for running Psspy case simulations."""
    def __init__(self, output_name, input_network="Scenario1"):
        """
            Initializes the object for PSSE simulations.
            Input:
                output_name: Name of output-file which simulation data will be saved.
                input_network: Name of .sav and .dyr files which will be loaded from a folder.
        """
    # Store information about the case in the object
        self.output_name = output_name  # Name of output-file which simulation data will be saved.
        self.input_network = input_network  # Name of .sav and .dyr files which will be loaded from a folder.
        self.hvdc_bus_nrs = []  # List of buses that are HVDC?
        self.hvdc_limits = []  # Limits for HVDC?
        self.filename = ""  # The filename used for the case at hand?
        self.sh_ttl = ""  # Title? From .out-files
        self.ch_id = {}  # Channel identifiers corresponding to data output from .out-files
        self.ch_data = {}  # Data output from .out-files
        self.psse_time_step = 0.01  # The time step for simulations in PSS/E
        self.end_time = 10.0  # Final length of simulation time in seconds [s]
        self.d_print = 100  # Setting for PSS/E, how many time steps between each write to console
        self.d_write = 10  # Setting for PSS/E, how many time steps between each write to output folder. Also is the
        # amount of time-steps between each stop to manually check values and initiate control actions
        self.machine_monitor_overview = 0  # For monitoring machine quantities
        self.bus_monitor_overview = 0  # For monitoring bus quantities
        self.present_plots = False  # True if plots are to be shown to user, False if should be directly saved without first being shown
        self.save_plots_local_folder = False  # True if plots are to be saved in the working folder, False to save in Dropbox-folder synced with Sharelatex
        self.events_overview = []  # For events (faults and disturbances) in dynamic analysis, # 1st column time at
        # which event occurs, 2nd column type of event (Branch fault? Load change?), 3rd param bus nr.
        # See self._exec_fault()

        self.machines_to_monitor = []
        self.buses_to_monitor = []

        # For calculating inertia
        self.contingency_time = []  # The time at which a contingency occurs
        self.contingency_size = []  # The size of the load step that occurs (load step emulates outage)
        self.total_initial_outage_size = 0  # The size of the initial outage in MW, necessary for inertia calculation

        # Other options
        self.branch_p_channels = []  # Used for keeping track of which channel indices correspond to which branch flows.
        self.calculated_coi_frequency = False  # Indicates whether COI-frequency has been calculated (for plots)

        # Stopping dynamic simulation to run checks
        self.monitor_data_manually = False  # If True, stops dynamic simulation every few time steps and monitors key
        # values
        self.programmable_generation = []  # List used for storing objects for user-defined programmable generation
        self.coi_frequency_should_be_calculated = False  # Decides whether COI-frequency needs to be calculated and thus
        # whether all machine channels should be monitored
        self.pgu_should_be_used = False
        self.bool_skip_pgu_updates_after_x_seconds = False  # If True, PGU updates are skipped after x seconds in
        # simulation time (specified in float_skip_pgu_updates_after_x_seconds
        self.float_skip_pgu_updates_after_x_seconds = 10.0  # Skip PGU updates after x seconds

        # PSS/E and Psspy setup
        redirect.psse2py()  # Redirect the PSS/E output to the terminal
        psspy.throwPsseExceptions = True

        # Files and folders
        os.chdir("..")
        cwd = os.getcwd()  # Get the current directory
        models = os.path.join(cwd, "Models")  # Name of the folder with the models
        self.case_file = os.path.join(models, input_network + ".sav")  # Static network data
        self.dyr_file = os.path.join(models, input_network + ".dyr")  # Dynamic model parameters

        # Initialize case
        psspy.psseinit(10000)
        psspy.case(self.case_file)  # Read in the power flow data
        psspy.dyre_new([1, 1, 1, 1], self.dyr_file, "", "", "")

    # ========================= Functions for case setup ===============================================================
    def configure_case_options(self, monitor_branch_flows=False,
                               end_time=10.0, d_print=100, d_write=10, present_plots=True,
                               save_plots_local_folder=False, coi_frequency_should_be_calculated=True,
                               pgu_should_be_used=True, bool_skip_pgu_updates_after_x_seconds=False,
                               float_skip_pgu_updates_after_x_seconds = 10.0):
        """
            Configures options for running the case study.
            Input:
                run_with_ufls: Determines whether Under-Frequency Load Shedding will be active.
                ufls_freq_threshold: Sets the frequency threshold for activation of UFLS.
                run_with_uvls: Determines whether Under-Voltage Load Shedding will be active.
                uvls_volt_threshold: Sets the voltage threshold for activation of UVLS.
                monitor_branch_flows:
                end_time: Total length in seconds [s] of time-domain simulation
                d_print: Time steps at which the solver prints
            Output:
                - Blank -
        """

        # Sets object booleans for overview
        self.monitor_branch_flows = monitor_branch_flows
        self.present_plots = present_plots
        self.save_plots_local_folder = save_plots_local_folder

        self.end_time = end_time
        self.d_print = d_print
        self.d_write = d_write

        self.coi_frequency_should_be_calculated = coi_frequency_should_be_calculated
        self.pgu_should_be_used = pgu_should_be_used
        self.bool_skip_pgu_updates_after_x_seconds = bool_skip_pgu_updates_after_x_seconds
        self.float_skip_pgu_updates_after_x_seconds = float_skip_pgu_updates_after_x_seconds

    def monitor_channels(self, machines_to_monitor=(5600, 3300, 7000), machine_quantities=(1, 2, 4, 7),
                         buses_to_monitor=(3000, 3100, 3249), bus_quantities=(1, 2, 4), branch_from_to=[[],[]],
                         loads_to_monitor_power=[3000, 3100, 3249]):
        """
            Sets which channels should be monitored in the dynamic simulation and saved in the output-file.
            Input:
                machines_to_monitor: Which generator buses should be monitored.
                machine_quantities: Which channels at generator buses that should be monitored.
                buses_to_monitor: Which load buses should be monitored.
                bus_quantities: Which quantities at the load buses that should be monitored.
                branch_from_to:A list containing two lists: the first list contains one end of the branch that should be
                    monitored (the "from"-end or source end) and the second list contains the other end of the branches
                    that should be monitored. Input to each list should be bus numbers and lists must be equal.
                loads_to_monitor_power: A list of bus numbers where load active power is monitored
            Output:
                - BLANK -
        """
        self.machines_to_monitor = machines_to_monitor
        self.buses_to_monitor = buses_to_monitor

        chan_index = 1  # Used for identifying channel indices

    # Monitor machine quantities input by user
        machine_monitor_overview = []
        for j in range(len(machine_quantities)):
            if self.coi_frequency_should_be_calculated and machine_quantities[j] == 7:  # All machines for COI-frequency
                temp_machines_to_monitor = machines_to_monitor[:]
                machines_to_monitor = psspy.agenbusint(-1, 1, "NUMBER")[1][0]
            for i in range(len(machines_to_monitor)):
                # Add channels to monitor
                psspy.machine_array_channel([chan_index, machine_quantities[j], machines_to_monitor[i]])

                # Store for plotting later on
                new_row = [chan_index]
                new_row.append(int(machine_quantities[j]))  # "ANGLE" = 1, "PELEC" = 2 etc.
                new_row.append(int(machines_to_monitor[i]))  # Bus number
                machine_monitor_overview.append(new_row)  # Appends new row new_row

                chan_index = chan_index + 1  # len(machine_quantities)*i+j+1 could here be set equal to chan_index

            if self.coi_frequency_should_be_calculated and machine_quantities[j] == 7:  # After all machine frequencies
                # are set to be monitored, only monitor the machine quantities that were intended to be modified
                machines_to_monitor = temp_machines_to_monitor # As assigned above

        self.machine_monitor_overview = np.array(machine_monitor_overview)

    # Monitor load quantities
        bus_monitor_overview = []
        for j in range(len(bus_quantities)):

            if self.pgu_should_be_used == True and bus_quantities[j] == 1:
                temp_buses_to_monitor = buses_to_monitor[:]
                buses_to_monitor = psspy.alodbusint(-1, 1, "NUMBER")[1][0]

            for i in range(len(buses_to_monitor)):
                # Add channels to monitor
                if bus_quantities[j] == 1:  # Bus frequency channel
                    psspy.bus_frequency_channel([chan_index, buses_to_monitor[i]])
                elif bus_quantities[j] == 2:  # Bus active power load channel
                    psspy.load_array_channel([chan_index, 1, buses_to_monitor[i]])# Not yet implemented, remember to add all loads at the bus
                elif bus_quantities[j] == 3:  # Bus reactive power load
                    psspy.load_array_channel([chan_index, 2, buses_to_monitor[i]])  # Not yet implemented
                elif bus_quantities[j] == 4:  # Bus voltage
                    psspy.voltage_channel([chan_index, -1, -1, buses_to_monitor[i]])
                elif bus_quantities[j] == 5:  # Bus voltage angle
                    pass  # Not yet implemented
                else:
                    pass  # Not yet implemented

                # Store for plotting later on
                new_row = [chan_index]
                new_row.append(int(bus_quantities[j]))
                new_row.append(int(buses_to_monitor[i]))
                bus_monitor_overview.append(new_row)
                chan_index = chan_index + 1  # len(bus_quantities)*i+j+1 could here be set equal to chan_index????

            if self.pgu_should_be_used == True and bus_quantities[j] == 1:  # Return to the previous list of
                # buses to monitor
                buses_to_monitor = temp_buses_to_monitor

        self.bus_monitor_overview = np.array(bus_monitor_overview)


    # Monitor branch flows
        if len(branch_from_to) > 0 and len(branch_from_to[0]) == len(branch_from_to[1]):
            for i in range(len(branch_from_to[0])):
                psspy.branch_p_channel([_i, _i, _i, branch_from_to[0][i], branch_from_to[1][i]], _s, _s)  # Test this shit
                self.branch_p_channels.append([chan_index, branch_from_to[0][i], branch_from_to[1][i]])
                chan_index = chan_index + 1
        else:  # If input from_to_branch is not formatted properly
            self.error


    # Also add bus voltage monitoring for all buses as this will be needed for the loads to be able to react to voltage
    # quantities

    # Also add code that adds frequency monitoring for the buses that have PGUs

# PROGRAMMABLE GENERATION UNITS

    def add_programmable_generation(self, gen_type="DEFAULT_BATTERY", bus_number=5600):
        """
            Adds a programmable generation unit (PGU) to the specified bus. The response scheme of the PGU can be set by
             the user by altering or adding the variables found under the different gen_type options. The PGU works by
             updating the PGU every few milliseconds in dynamic simulation time with the latest simulation data, and
             then checking whether activation conditions are met.

             PGUs MUST BE ADDED BEFORE MONITOR_CHANNELS
            Input:
                bus_number: The bus number where the PGU is to be placed.
                gen_type: An option for which kind of PGU the user desired to have implemented at the bus, to easily
                    replicate the PGUs between cases.
            Output:
                - Blank -
        """
        self.monitor_data_manually = True  # If a PGU is added, the system will have to stop dynamic simulation every
        # few milliseconds

        limits = []  # Activation limits in frequency [Hz], ROCOF [Hz/s] or voltage (p.u.)
        activation_at_limits = []  # How many MW in power should be activated for each of the limits reached
        time_to_activate = []  # How long it should take for the load step to kick in (approximately)
        max_time_available = []  # How long the unit is capable of supplying the individual load step (approximately)
        max_output = 10000  # The maximum power output of this generation unit

        if gen_type == "DEFAULT_BSFREQ":
            monitor_type = "BSFREQ"
            limits = [49.9, 49.8, 49.7, 49.6, 49.5]
            activation_at_limits = [300, 300, 300, 300, 300]
            time_to_activate = [1, 1, 1, 1, 1]
            max_time_available = [50.0, 5.0, 5.0, 5.0, 5.0]
            max_output = 3000  # MW
        elif gen_type == "pgu_ufls_interaction_test":
            monitor_type = "BSFREQ"
            limits = [49.9, 49.8, 49.7, 49.6, 49.5]
            activation_at_limits = [150, 150, 150, 150, 150]
            time_to_activate = [1.0, 1.0, 1.0, 1.0, 1.0]
            max_time_available = [3.0, 3.0, 3.0, 3.0, 3.0]
            max_output = 1000  # MW
        elif gen_type == "debugging":
            monitor_type = "BSFREQ"
            limits = [49.9, 49.8, 49.7, 49.6, 49.5]
            activation_at_limits = [400, 400, 400, 400, 400]
            time_to_activate = [1.0, 1.0, 1.0, 1.0, 1.0]
            max_time_available = [3.0, 3.0, 3.0, 3.0, 3.0]
            max_output = 1000  # MW
        elif gen_type == "test_remains_active":
            monitor_type = "BSFREQ"
            limits = [49.5]
            activation_at_limits = [1500]
            time_to_activate = [1.0]
            max_time_available = [3.0]
            max_output = 2000  # MW
        elif gen_type == "NORNED_HVDC_EM_POWER":
            bus_number = 5620
            monitor_type = "BSFREQ"
            limits = [49.4]
            activation_at_limits = [140]
            time_to_activate = [0.5]
            max_time_available = [90.0]
        elif gen_type == "SKAGERAK_1-4_HVDC_EM_POWER":
            bus_number = 5610
            monitor_type = "BSFREQ"
            limits = [49.5]
            activation_at_limits = [340]
            time_to_activate = [0.5]
            max_time_available = [90.0]
        elif gen_type == "KONTISKAN_1-2_HVDC_EM_POWER":
            bus_number = 3360
            monitor_type = "BSFREQ"
            limits = [49.2]
            activation_at_limits = [148]
            time_to_activate = [0.5]
            max_time_available = [90.0]
        elif gen_type == "BALTIC_HVDC_EM_POWER":
            bus_number = 8600
            monitor_type = "BSFREQ"
            limits = [49.3]
            activation_at_limits = [120]
            time_to_activate = [0.5]
            max_time_available = [90.0]
        elif gen_type == "SWEPOL_HVDC_EM_POWER":
            bus_number = 8700
            monitor_type = "BSFREQ"
            limits = [49.4]
            activation_at_limits = [120]
            time_to_activate = [0.2]
            max_time_available = [90.0]
        elif gen_type == "NORDBALT_HVDC_EM_POWER":
            # bus_number = XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
            monitor_type = "BSFREQ"
            limits = [49.5]
            activation_at_limits = [140]
            time_to_activate = [0.5]
            max_time_available = [90.0]
        elif gen_type == "FENNOSKAN_HVDC_EM_POWER":
            # bus_number = XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
            monitor_type = "BSFREQ"
            limits = [49.5]
            activation_at_limits = [100]
            time_to_activate = [0.5]
            max_time_available = [90.0]
        elif gen_type == "ESTLINK_1-2_HVDC_EM_POWER":
            bus_number = 7020
            monitor_type = "BSFREQ"
            limits = [49.6]
            activation_at_limits = [200]
            time_to_activate = [0.5]
            max_time_available = [90.0]
        elif gen_type == "VYBORG_HVDC_EM_POWER":
            bus_number = 7010
            monitor_type = "BSFREQ"
            limits = [49.7]
            activation_at_limits = [284]
            time_to_activate = [0.5]
            max_time_available = [90.0]
        elif gen_type == "NORTH_SEA_LINK_HVDC_EM_POWER":
            bus_number = 6010
            monitor_type = "BSFREQ"
            limits = [49.5]
            activation_at_limits = [280]
            time_to_activate = [0.2]
            max_time_available = [90.0]
        elif gen_type == "NORDLINK_HVDC_EM_POWER":
            bus_number = 5630
            monitor_type = "BSFREQ"
            limits = [49.5]
            activation_at_limits = [280]
            time_to_activate = [0.5]
            max_time_available = [90.0]
        else:
            self.force_error += 1  # This line forces an error

        p_gen = pg.ProgrammableGenerationUnit(bus_number, monitor_type, self.d_write, limits, activation_at_limits,
                                          time_to_activate, max_time_available, max_output, True)
        self.programmable_generation.append(p_gen)

    def add_demand_response(self, focl=0.001, time_delay=0.01, linear=True, resolution=100, ):
        """
            Adds demand response to all buses in the system by adding programmable generation unit to each load. The
            output of the PGU is scaled by the load by a percentage focl.

            PGUs MUST BE ADDED BEFORE MONITOR_CHANNELS
        :param focl: Fraction of controllable load, for example 0.1 if 10% of loads can be controlled / disconnected
        :param activation_param: the threshold for which
        :return:
        """
        self.monitor_data_manually = True  # If a demand response is added, the system will have to stop dynamic
        # simulation every few milliseconds
        start_activate_dsr_frequency = 49.9  # DSR starts activating as frequency falls below 49.9 Hz
        all_dsr_active_frequency = 49.0  # All DSR activated at 49.0 Hz


        # Find all loads
        load_bus_numbers_references = np.array(psspy.alodbusint(-1, 1, "NUMBER")[1][0])  # The load numbers
        loads_mva = np.array(psspy.alodbuscplx(-1, 1, "TOTALACT")[1][0])  # The loads in MVA
        loads_mva = np.real(loads_mva)

        debugging = True
        if debugging:
            #load_bus_numbers = np.array([3100, 3300, 5500, 5600, 7000])  # WORKS
            #load_bus_numbers = np.array([3000, 3100, 3300, 5500, 5600, 7000])  # DOESN'T WORK, LOAD 3000 GOES CRAZY
            load_bus_numbers = np.array([3100, 3115, 3249, 3300, 3359, 5100, 5300, 5400, 5500, 5600, 6100, 6500, 6700, 7000, 7100, 8500])
            #load_bus_numbers = np.array([3000, 3100, 3115, 3249, 3300, 3359, 5100, 5300, 5400, 5500, 5600, 6100, 6500, 6700, 7000, 7100, 8500])

        load_nr = 0
        for i in range(len(load_bus_numbers)):
            current_load = loads_mva[i]  # The load to use to set parameters for programmable generation
            load_nr = load_bus_numbers[i]

            limits = np.linspace(start_activate_dsr_frequency, all_dsr_active_frequency, resolution)
            activation_at_limits = np.linspace(current_load*focl/resolution, current_load*focl/resolution, resolution)
            time_to_activate = np.linspace(time_delay, time_delay, resolution)
            max_time_available = np.linspace(15.0, 15.0, resolution)
            max_output = sum(activation_at_limits)

            # Finally, add PGU to the system
            p_gen = pg.ProgrammableGenerationUnit(load_nr, "BSFREQ", self.d_write, limits, activation_at_limits,
                                                  time_to_activate, max_time_available, max_output, True)
            self.programmable_generation.append(p_gen)

    def add_ufls(self, ufls_buses=[[]], ufls_settings=[[]]):
        """
            Adds UFLS protection relays to ALL or specified load buses
            Input:
                ufls_buses: specified as [[]] if all buses are to have ufls. If specific buses or sets of buses
                need to have different settings, sets of buses with similar settings can be seen separated by the outer list,
                and individual buses in the set by the inner list.
                ufls_settings: specified as [[]] if all buses are to have similar ufls_settings (for which the settings
                are specified in the code itself). Outer list again separates sets and inner list gives the settings for
                each of the ufls protection relays in the set. Outer list dimensions must be the same as outer list in
                ufls_buses as each set of buses needs specific settings.
            Output:
                - blank -
        """
        ufls_buses=[[]]  # TEMPORARY WORKAROUND AS ufls_buses  SEEMS TO BE MAGICALLY PASSED ON BETWEEN DSR SWEEPING ITERATIONS?
        if len(ufls_buses[0]) < 1:  # None specified, use ALL loads and load IDs
            ufls_buses[0] = np.array(psspy.aloadint(-1, 1, "NUMBER")[1][0])
            ufls_settings = [[48.6, 0.062, 48.2, 0.156, 47.9, 0.093]]

        for k in range(len(ufls_buses)):
            # Modify ufls_buses so it reflects the number of loads at each bus
            loads_at_buses = np.array(psspy.aloadint(-1, 1, "NUMBER")[1][0])
            unique_ufls_buses = np.unique(ufls_buses[k])
            extended_ufls_buses = []
            for i in range(len(unique_ufls_buses)):
                nr_of_loads_at_bus = len(np.where(loads_at_buses == unique_ufls_buses[i])[0])  # Number of loads at the bus nr, for example 3 loads
                #  at bus 3300
                for j in range(nr_of_loads_at_bus):
                    extended_ufls_buses.append(unique_ufls_buses[i])
            ufls_buses[k] = extended_ufls_buses

        # For N44 model with similar settings for each load bus:
        ufls_buses[0] = [3000, 3000, 3000, 3100, 3115, 3249, 3300, 3300, 3359, 3359, 3359, 3359, 5100, 5300, 5400, 5500,
                         5500, 5600, 5600, 6100, 6100, 6500, 6500, 6500, 6700, 7000, 7000, 7000, 7000, 7000, 7100, 7100,
                         8500, 8500, 8500]


        for k in range(len(ufls_buses)):
            # Add protection relay for each load and load ID
            load_nr = 0  # needs both number and ID
            load_id = 0  # which complicates the iteration through the loads
            for i in range(len(ufls_buses[k])):
                if load_nr != ufls_buses[k][i]:  # new load
                    load_id = 1
                    load_nr = ufls_buses[k][i]
                else:  # Same machine as previous
                    load_id = load_id + 1
                psspy.add_load_model(load_nr, str(load_id), 0, 2, r"""DLSHBL""", 0, [], [], 13, [ufls_settings[k][0],
                                                                                                 0.0,
                                                                                                 ufls_settings[k][1],
                                                                                                 ufls_settings[k][2],
                                                                                                 0.0,
                                                                                                 ufls_settings[k][3],
                                                                                                 ufls_settings[k][4],
                                                                                                 0.0,
                                                                                                 ufls_settings[k][5],
                                                                                                 0.0, 0.0, 0.0, 0.0])
                # Configure settings for each relay. Possibly add ROCOF-settings as well?

    def add_uvls(self, uvls_buses=[[]], uvls_settings=[[]]):
        """
            Adds UVLS protection relays to ALL or specified load buses and sets the correct settings
            Input:
                uvls_buses: specified as [[]] if all buses are to have uvls. If specific buses or sets of buses
                need to have different settings, sets of buses with similar settings can be seen separated by the outer list,
                and individual buses in the set by the inner list.
                uvls_settings: specified as [[]] if all buses are to have similar uvls_settings (for which the settings
                are specified in the code itself). Outer list again separates sets and inner list gives the settings for
                each of the uvls protection relays in the set. Outer list dimensions must be the same as outer list in
                uvls_buses as each set of buses needs specific settings.
            Output:
                - blank -
        """

        if len(uvls_buses[0]) < 1:  # None specified, use ALL loads and load IDs
            uvls_buses[0] = np.array(psspy.aloadint(-1, 1, "NUMBER")[1][0])
            uvls_settings = [[0.9, 0.1, 0.8, 0.1, 0.7, 0.1]]  # Default (arbitrary) settings. Shed 10% load at 0.9 p.u. voltage

        for k in range(len(uvls_buses)):
            # Modify uvls_buses so it reflects the number of loads at each bus
            loads_at_buses = np.array(psspy.aloadint(-1, 1, "NUMBER")[1][0])
            unique_uvls_buses = np.unique(uvls_buses[k])
            extended_uvls_buses = []
            for i in range(len(unique_uvls_buses)):
                nr_of_loads_at_bus = len(np.where(loads_at_buses == unique_uvls_buses[i])[
                                             0])  # Number of loads at the bus nr, for example 3 loads
                #  at bus 3300
                for j in range(nr_of_loads_at_bus):
                    extended_uvls_buses.append(unique_uvls_buses[i])
            uvls_buses[k] = extended_uvls_buses

        # For N44 model with similar settings for each load bus:
        # Only add UVLS in Sweden
        uvls_buses[0] = [3000, 3000, 3000, 3100, 3115, 3249, 3300, 3300, 3359, 3359, 3359, 3359, 8500, 8500, 8500]

        for k in range(len(uvls_buses)):
            # Add protection relay for each load and load ID
            load_nr = 0  # needs both number and ID
            load_id = 0  # which complicates the iteration through the loads
            for i in range(len(uvls_buses[k])):
                if load_nr != uvls_buses[k][i]:  # new load
                    load_id = 1
                    load_nr = uvls_buses[k][i]
                else:  # Same machine as previous
                    load_id = load_id + 1
                psspy.add_load_model(load_nr, str(load_id), 0, 2, r"""LVSHBL""", 1, [0], [""], 10,
                                     [uvls_settings[k][0], 0.0, uvls_settings[k][1], uvls_settings[k][2], 0.0,
                                      uvls_settings[k][3], uvls_settings[k][4], 0.0, uvls_settings[k][5], 0.0])


    def add_hvdc_buses(self):
        """
            Adds the HVDC-links North Sea Link and NordLink to the Nordic 44 model. The modelling is the same as
            previous HVDC-links that have been implemented in the model.
        """

        # North Sea Link nsl ( http://www.statnett.no/en/Projects/Cable-to-the-UK/ )
        nsl_nr = 6010
        nsl_area = 14
        nsl_connection = 6000  # Kvilldal
        nsl_voltage = 300.0
        nsl_reactance = 0.006
        nsl_initial_power = 0.0
        nsl_bus_type = 2  # 1 = PQ, 2 = PV? What is -2?

        psspy.bus_data_3(nsl_nr, [_i, nsl_area, _i, _i], [nsl_voltage, _f, _f, _f, _f, _f, _f], _s)
        psspy.load_data_4(nsl_nr, r"""1""", [_i, _i, _i, _i, _i, _i], [_f, _f, _f, _f, _f, _f])  # Create bus before any
        #  values can be set
        psspy.load_chng_4(nsl_nr, r"""1""", [_i, _i, _i, _i, _i, _i], [nsl_initial_power, _f, _f, _f, _f, _f])  # Set
        # values (power)
        psspy.branch_data(nsl_connection, nsl_nr, r"""1""", [_i, _i, _i, _i, _i, _i],
                          [_f, _f, _f, _f, _f, _f, _f, _f, _f, _f, _f, _f, _f, _f, _f])  # Create branch before setting
        # values
        psspy.branch_chng(nsl_connection, nsl_nr, r"""1""", [_i, _i, _i, _i, _i, _i],
                          [_f, nsl_reactance, _f, _f, _f, _f, _f, _f, _f, _f, _f, _f, _f, _f, _f])  # Set values
        # (reactance)

        # NordLink nl ( http://www.statnett.no/Nettutvikling/NORDLINK/ )
        nl_nr = 5630
        nl_area = 13
        nl_connection = 5600  # Kristiansand?
        nl_voltage = 300.0
        nl_reactance = 0.006
        nl_initial_power = 0.0
        nl_bus_type = 2  # 1 = PQ, 2 = PV? What is -2?

        psspy.bus_data_3(nl_nr, [_i, nl_area, _i, _i], [nl_voltage, _f, _f, _f, _f, _f, _f], _s)
        psspy.load_data_4(nl_nr, r"""1""", [_i, _i, _i, _i, _i, _i], [_f, _f, _f, _f, _f, _f])  # Create bus before any
        # values can be set
        psspy.load_chng_4(nl_nr, r"""1""", [_i, _i, _i, _i, _i, _i], [nl_initial_power, _f, _f, _f, _f, _f])  # Set
        # values (power)
        psspy.branch_data(nl_connection, nl_nr, r"""1""", [_i, _i, _i, _i, _i, _i],
                          [_f, _f, _f, _f, _f, _f, _f, _f, _f, _f, _f, _f, _f, _f, _f])  # Create branch before setting
        # values
        psspy.branch_chng(nl_connection, nl_nr, r"""1""", [_i, _i, _i, _i, _i, _i],
                          [_f, nl_reactance, _f, _f, _f, _f, _f, _f, _f, _f, _f, _f, _f, _f, _f])  # Set values
        # (reactance)

    def add_event(self, time, type, bus, extra_params):
        """
            Adds an event to the system time-domain simulation. Events may include bus faults, load changes or branch
            faults, among others. See _exec_event for execution of these events. The code should be easily adaptable to
            include different kinds of events in dynamic simulations.
            Input:
                time: the time in seconds [s] at which the event occurs
                type: the type of event that occurs. 1=branch trip, 2=load step, 3=bus trip
                extra_params: possible extra required parameters. Bus number for bus trip, branch start and end for
                branch trip, load change and load bus number for load step, etc. See the respective functions in
                _exec_event for which exact required parameters.
        """
        # Input param extras will be read differently depending on type
        # Example: time=0.1, type=1, bus=5600, extra_params=(6000) to trip branch connecting 5600 and 6000
        if type == 1 or 2 or 3:  # Load step
            self.events_overview.append((time, type, bus, extra_params))
        else:
            return "Faulty argument type given"

        if type == 2:  # Load step, count size and time of outages set by user
            self.contingency_time.append(time)
            self.contingency_size.append(extra_params[0])

    # ========================= Functions for dynamic simulations ======================================================
    def prepare_dynamic_simulation(self, psse_time_step=0.005, p_zip=[10.0, 10.0], q_zip=[10.0, 10.0]):
        """
            Prepares PSSE for dynamic simulations by setting output-file paths, converting loads to ZIP-model and setting
            solution parameters.
        """
        self.output_file = os.path.join(os.getcwd(), self.filename)  # Store dynamic simulation
        self.psse_time_step = psse_time_step

        # Convert the loads for dynamic simulation
        psspy.cong(0)
        psspy.conl(0, 1, 1, [0, 0], [p_zip[0], p_zip[1], q_zip[0], q_zip[1]])
        psspy.conl(0, 1, 2, [0, 0], [p_zip[0], p_zip[1], q_zip[0], q_zip[1]])
        psspy.conl(0, 1, 3, [0, 0], [p_zip[0], p_zip[1], q_zip[0], q_zip[1]])  # Tuning from S. M. Hamre's thesis

        # Set the time step for the dynamic simulation
        psspy.dynamics_solution_params(realar=[_f, _f, psse_time_step, _f, _f, _f, _f, _f])

        # Enable relative angle monitoring
        ibusex = 0  # 0 for system average, 3300 for setting 3300 as reference
        psspy.set_relang(1, ibusex)


    def run_dynamic_simulation(self):
        """
            Runs the time-domain simulation to the end, stopping to perform user-defined events at specific times
            (disconnection of branches, bus faults, or other events as programmed).
        """
        # Initialize
        ierr = psspy.strt(0, self.output_file)  # Tell PSS/E to write to the output file

        # Run and perform scheduled faults in queue
        # First need the initial run
        time = self.events_overview[0][0]  # The first event
        psspy.run(0, time, self.d_print, self.d_write, 0)  # Run dynamic simulation until the first event occurs

        # Now iterate through all faults
        time = self._find_unique_times()  # Returns a numpy array

        if self.monitor_data_manually:  # If I wish to stop the simulation every x time steps to perform control actions
            next_time = time[0]
            for i in range(len(time)):
                while next_time <= time[i]:  # Between events happening at different times
                    next_time = next_time + self.d_write * self.psse_time_step
                    self._supervise_system()
                    psspy.run(0, next_time, self.d_print, self.d_write, 0)
                indices = np.where(time == time[i])[0]  # Finds which events are to happen at present time
                self._exec_event(indices)  # Executes events given at the indices

                while next_time <= self.end_time:  # After last event and till the end
                    next_time = next_time + self.d_write * self.psse_time_step
                    self._supervise_system()
                    if self.bool_skip_pgu_updates_after_x_seconds:  # For shortening computation time when PGU does not
                        # change output after x seconds anyway
                        if next_time > self.float_skip_pgu_updates_after_x_seconds:
                            next_time = (self.end_time + 1)  # +1 to avoid floating point errors
                    psspy.run(0, next_time, self.d_print, self.d_write, 0)

        else:  # If I do not need to stop simulations and can rely entirely on built-in PSS/E features
            for i in range(len(time)):
                indices = np.where(time == time[i])[0]  # Finds which events are to happen at present time
                self._exec_event(indices)  # Executes events given at the indices
                if (i + 1) < len(time):  # Avoid going out of bounds, +1 due to index 0 is for chnf format
                    psspy.run(0, time[i + 1], self.d_print, self.d_write, 0)

            # At the last fault, run till end_time (the end of the simulation).
            psspy.run(0, self.end_time, self.d_print, self.d_write, 0)

    def _exec_event(self, indices):
        """
            Executes the faults in dynamic simulation that the user has set to happen for the given case study. See
            add_fault for details on executable faults.
            Input:
                indices: indices of which faults to execute in the events_overview variable of the object
        """
        # Index of fault in events_overview
        # indices = indices of faults to execute (potentially multiple at once)
        # Remember for the run_dynamic_simulation to use np.unique to decide for each time
        for i in range(len(indices)):
            type_of_fault = self.events_overview[indices[i]][1]  # Type of fault at the specified index

            if type_of_fault == 1:  # Branch trip
                self._exec_branch_trip(indices[i])
            elif type_of_fault == 2:  # Steps P at load and sets Q to zero.
                self._exec_load_step_and_set_q_to_zero(indices[i])
            elif type_of_fault == 3:  # Trips/disconnects a bus
                self._exec_bus_trip(indices[i])
            else:
                raise Exception("Undefined fault type, can not execute")

    def _exec_branch_trip(self, index):
        """
            Trips a branch.
            Input:
                index: index of the fault in self.events_overview that is to be executed
            Output:
                - Blank -
        """

        # Function not yet tested; not yet run any case studies with branch tripping
        time = self.events_overview[index][0]
        bus = self.events_overview[index][2]

        other_end = self.events_overview[index][3][0]
        branch_start = min(bus, other_end)
        branch_end = max(bus, other_end)
        psspy.dist_branch_trip(branch_start, branch_end, 1)  # Last integer assumes branch ID 1!!!!

    def _exec_bus_trip(self, index):
        """
            Trips/disconnects a bus.
            Input:
                index: index of the faults in self.events_overview that is to be executed.
            Output:
                - Blank -
        """

        # Function not yet tested; not yet run any case studies with bus tripping
        bus_number = self.events_overview[index][2]
        psspy.dist_bus_trip(bus_number)

    def _exec_load_step_and_set_q_to_zero(self, index):
        """
            Steps P at a load and sets Q to ZERO
            Input:
                index: index of the faults in self.events_overview that is to be executed.
            Output:
                - Blank -
        """

        # Applies a step load (MW) AND sets REACTIVE POWER to ZERO at the specified time (in seconds)
        bus_number = self.events_overview[index][2]
        load_step = self.events_overview[index][3][
            0]  # Third index refers to list of 'extra_params' for the specific
        # type of fault

        load_numbers = np.array(psspy.aloadint(-1, 1, "NUMBER")[1][0])
        load_index = np.where(load_numbers == bus_number)[0]
        load_step = load_step / len(load_index)  # To split step over all loads at bus

        load_p = np.array(psspy.aloadcplx(-1, 1, "MVAACT")[1][0])
        present_load = load_p[load_index]
        for i in range(len(load_index)):  # CREATE FOR LOOP NOW TO ITERATE THROUGH AND STEP LOAD AT EACH LOAD AT BUS
            load_id = str(i + 1)
            #psspy.load_chng_4(bus_number, load_id, [_i, _i, _i, _i, _i, _i], [present_load[i] + load_step, 0.0,
                                                                              #_f, _f, _f, _f])
            psspy.load_chng_4(bus_number, load_id, [_i, _i, _i, _i, _i, _i], [0.0, 0.0, 0.0, 0.0, 0.0, 0.0])
            # NB!! The present_load seems to take values from the Machines-tab

    def disable_primary_response(self):
        # Loop through all dynamic data and set all HYGOV governors out of service. Only necessary to put HYGOV out of
        # service as the hydro turbines are the only ones contributing to primary response
        all_machine_numbers = np.array(psspy.amachint(-1, 4, "NUMBER")[1][0])  # Note that the second input parameter;
                                # retrieves all machine numbers, not just online machines
        #all_machine_numbers = psspy.amachint(-1, 4, "NUMBER")
        machine_nr = 0
        machine_id = 0
        for i in range(len(all_machine_numbers)):
            if machine_nr != all_machine_numbers[i]:  # New plant
                machine_id = 1
                machine_nr = all_machine_numbers[i]
            else:  # Same plant as previous machine
                machine_id += 1
            governor_name = psspy.mdlnam(machine_nr, str(machine_id), "GOV")[1]
            if governor_name[0:5] == "HYGOV":
                psspy.plmod_status(machine_nr, str(machine_id), 7, 0)

        breakpoint = 1


    def _supervise_system(self):
        """
            Supervises system by running various checks for each stop in the dynamic simulation.
            Function passes on bus frequency data to the PGUs so
            that they can respond by increasing their active power output.
        """

        # Acquire newest system data
        chnf = dyntools.CHNF(self.output_file + ".out")
        self.sh_ttl, self.ch_id, self.ch_data = chnf.get_data()

        # Adjusts the setpoints for the programmable generation in response to changing frequencies and voltages
        for i in range(len(self.programmable_generation)):
            bus_number = self.programmable_generation[i].return_bus_number()

            bus_number_match = np.where(self.bus_monitor_overview[:, 2] == bus_number)[0]
            bus_frequency_monitoring_match = np.where(self.bus_monitor_overview[:, 1] == 1)[0]
            for j in range(len(bus_number_match)):
                temp_index = np.where(bus_frequency_monitoring_match == bus_number_match[j])[0]
                if len(temp_index) > 0:
                    ch_index = self.bus_monitor_overview[bus_frequency_monitoring_match[temp_index[0]]][0]
                    break
            self.programmable_generation[i].update_pgu(self.ch_data["time"], self.ch_data[ch_index])

    # Functions for static calculations
    def set_hvdc_active_power_and_redistribute(self, hvdc_bus_nr=5610, hvdc_setpoint=1400):
        """
            Sets the active power at a HVDC load bus to a specified value, and redistributes the load change across the
            system to avoid significant change in the slack bus generation and total system power flows. This is done
            because the system has been tuned for a case study in S. M. Hamre's thesis, and it is attempted to retain
            the tuning valid. An example: if cable A is currently importing 100 MW, but its capacity is 200 MW, the load
            consumption may be changed to -200 MW, for which an increase of (-100-(-200)) = 100 MW in generation is seen
            by the system. By reducing the system-wide generation, the slack bus generation is not changed by 100 MW but
            the system instead absorbs the change.
            Input:
                hvdc_bus_nr: The HVDC bus which should see its power changed.
                hvdc_setpoint: The setpoint of the HVDC link, typically set to the link's maximum transfer capacity.
            Output:
                - Blank -
        """
        # Define bus number and read load at this bus
        self.hvdc_bus_nrs.append(hvdc_bus_nr)  # PSS/E bus number
        self.hvdc_limits.append(hvdc_setpoint)  # Maximum capacity of this HVDC link
        self.filename = self.output_name

        # Read the load at the HVDC bus
        load_p = np.array(psspy.aloadcplx(-1, 1, "MVAACT")[1][
                             0])  # Actual MVA (MVAACT) and Nominal MVA (MVANOM) seem to give same value
        load_numbers = np.array(psspy.aloadint(-1, 1, "NUMBER")[1][0])  # Load all bus numbers
        load_p = load_p[
            np.where(load_numbers == hvdc_bus_nr)[0][0]]  # load_p now has the value of the load at hvdc bus before
        # adjustment

        # Set cable to max export
        psspy.load_chng_4(hvdc_bus_nr, _s, [_i, _i, _i, _i, _i, _i],
                          [hvdc_setpoint, _f, _f, _f, _f, _f])

        # The generation necessary for increased export or import at the cable is distributed over available capacity
        # in the system. This is so the slack bus doesn't have to adjust too much
        p_change_due_to_setting_hvdc_power = hvdc_setpoint - load_p  # Warning: load_p is a complex value
        p_change_due_to_setting_hvdc_power = p_change_due_to_setting_hvdc_power.real  # Convert to real value, no longer complex as

        # Area of the HVDC bus
        bus_numbers = np.array(psspy.abusint(-1, 1, "NUMBER")[1][0])  # Get all bus numbers
        bus_areas = np.array(psspy.abusint(-1, 1, "AREA")[1][0])
        hvdc_bus_area = bus_areas[
            np.where(bus_numbers == hvdc_bus_nr)[0][0]]  # Use bus numbers to index for the area of HVDC bus

        bool_increase_gen = p_change_due_to_setting_hvdc_power > 0.0  # Bool
        if bool_increase_gen:  # If the export of HVDC-link is increased instead of decreased
            raise  # We only wish to study the case of increased import in this thesis, not increase export

        else:  # Increased import at HVDC-link, as thesis is concerned about

            # Find generation capacity of this specific area
            areas_of_plants = np.array(psspy.agenbusint(-1, 1, "AREA")[1][0])
            indices = np.where(areas_of_plants == hvdc_bus_area)[0]
            gen_of_plants = np.array(psspy.agenbusreal(-1, 1, "PGEN")[1][0])  # Retrieve active power generation of all plants
            gen_in_hvdc_bus_area = gen_of_plants[indices]  # Only include area of hvdc bus
            gen_capacities_of_plants = np.array(psspy.agenbusreal(-1, 1, "PMAX")[1][0])  # Retrieve generation capacity of all plants
            gen_capacity_in_hvdc_bus_area = gen_capacities_of_plants[indices]
            sum_gen_in_hvdc_bus_area = sum(gen_in_hvdc_bus_area)  # Currently active generation in the HVDC area
            sum_gen_of_plants = sum(gen_of_plants)  # Currently active generation in the system

            # psspy.amachint can not return "AREA", thus we need a workaround using machine numbers and bus areas
            # Find "AREA" of each machine
            machine_numbers = np.array(psspy.amachint(-1, 1, "NUMBER")[1][0])
            machine_areas = np.zeros(len(machine_numbers))
            for i in range(len(machine_numbers)):  # For loop to assign area numbers
                machine_areas[i] = bus_areas[np.where(bus_numbers == machine_numbers[i])]

            indices = np.where(machine_areas == hvdc_bus_area)[0]  # Indices of machines in the HVDC bus area
            machine_generation = np.array(psspy.amachreal(-1, 1, "PGEN")[1][0])  # Read PGEN of machines
            machine_generation_capacity = np.array(psspy.amachreal(-1, 1, "PMAX")[1][0])  # Read PMAX of machines
            machine_gen_change = machine_generation  # The generation that can be reduced when import is increased. NB:
            # machine_generation includes ALL machines, not just the ones in HVDC area

            current_machine_nr = 0  # machine_chng_2 needs both number and ID
            machine_id = 0  # which complicates the iteration through the machines

            # Distribute generation change in local region
            if abs(sum_gen_in_hvdc_bus_area) > abs(p_change_due_to_setting_hvdc_power):  # Distribute gen change over local
                # area if the generators can take it all
                gen_change_ratio = abs(p_change_due_to_setting_hvdc_power) / sum_gen_in_hvdc_bus_area
                machine_gen_change = machine_gen_change* gen_change_ratio

                # Set each generator to unique value
                for i in indices:  # Iterate through each in-service machine and set generation
                    if current_machine_nr != machine_numbers[i]:  # new machine
                        machine_id = 1
                        current_machine_nr = machine_numbers[i]
                    else:  # Same machine as previous
                        machine_id = machine_id + 1

                    psspy.machine_chng_2(current_machine_nr, str(machine_id),
                                         [_i, _i, _i, _i, _i, _i],
                                         [machine_generation[i] - machine_gen_change[i], _f, _f, _f, _f, _f, _f,
                                          _f, _f, _f, _f, _f, _f, _f, _f,
                                          _f, _f])

            else:  # Distribute over whole system

                # First set all generation to 0 in local area
                for i in indices:  # Iterate through machines in local area and set to zero generation
                    if current_machine_nr != machine_numbers[i]:  # new machine
                        machine_id = 1
                        current_machine_nr = machine_numbers[i]
                    else:  # Same machine as previous
                        machine_id = machine_id + 1

                    psspy.machine_chng_2(current_machine_nr, str(machine_id),
                                         [_i, _i, _i, _i, _i, _i],
                                         [0.0, _f, _f, _f, _f, _f, _f, _f,
                                          _f, _f, _f, _f, _f, _f, _f, _f, _f])

                # After setting to 0 in local area, recalculate slack to distribute over system
                p_change_due_to_setting_hvdc_power = p_change_due_to_setting_hvdc_power + \
                                                     sum(machine_gen_change[indices])  # Recalculate the slack that
                # needs to be distributed
                system_remaining_generation_capacity = sum_gen_of_plants - sum(machine_gen_change[indices])  # Recalculate remaining generation capacity of
                # system
                machine_gen_change[indices] = 0  # Generators at 0 = can't take more slack
                gen_change_ratio = abs(p_change_due_to_setting_hvdc_power) / system_remaining_generation_capacity
                machine_gen_change = machine_gen_change * gen_change_ratio

                # Distribute this new slack across the rest of the system
                machine_id = 0
                current_machine_nr = 0
                for i in range(len(machine_gen_change)):  # Increase generation of all generators in system
                    if current_machine_nr != machine_numbers[i]:  # New machine
                        machine_id = 1
                        current_machine_nr = machine_numbers[i]
                    else:  # Same machine as previous
                        machine_id = machine_id + 1
                    psspy.machine_chng_2(current_machine_nr, str(machine_id),
                                         [_i, _i, _i, _i, _i, _i],
                                         [machine_generation[i] - machine_gen_change[i], _f, _f, _f, _f, _f, _f,
                                          _f, _f, _f, _f, _f, _f, _f, _f, _f, _f])
            self.reduce_to_minimum_generators_needed_for_generation()  # All generation has now been distributed

            # Now all slack has been distributed

    def reduce_to_minimum_generators_needed_for_generation(self):
        """
        Used to calculate how many machines are needed online for each plant to produce the amount of active power that
        currently being produced. If 4 machines are each running at 20% of maximum generation capacity, this function
        will set the first generator to 80% output and the other generators to zero output, without disconnecting them.
        The function attempts to minimize the generators online for supplying the same amount of active power.

        # Previously named "after_import_adjustment". Function is called in "set_hvdc_active_power_and_redistribute"
        :param hvdc_bus_area:
        :param machine_areas:
        :return:
        """



        # Adjust import for each plant in the area separately
        plant_areas = np.array(psspy.agenbusint(-1, 1, "AREA")[1][0])
        areas = np.unique(plant_areas)
        for k in range(len(areas)):  # Iterate through all areas
            plant_numbers = np.array(psspy.agenbusint(-1, 1, "NUMBER")[1][0])
            plant_numbers = plant_numbers[np.where(areas[k] == plant_areas)]  # Plant numbers in area

            machine_numbers = np.array(psspy.amachint(-1, 1, "NUMBER")[1][0])  #
            machine_gen = np.array(psspy.amachreal(-1, 1, "PGEN")[1][0])
            machine_max_gen = np.array(psspy.amachreal(-1, 1, "PMAX")[1][0])

            for i in range(len(plant_numbers)):  # Iterate through all plants in area
                plant_number = plant_numbers[i]
                indices = np.where(plant_number == machine_numbers)[0]
                machine_plant_gen = machine_gen[indices]  # Generation of machines at this specific plant
                machine_plant_max_gen = machine_max_gen[indices]  # Max generation of machines at this specific plant

                gen = sum(machine_plant_gen)  # Find total generation of this plant in this area
                max_gen = machine_plant_max_gen[0]  # Index 0 as all generators at the plant have same max generation
                nr_of_gens = int(np.ceil(gen / max_gen))  # The number of generators necessary to take generate this
                single_gen_value = gen / nr_of_gens  # Generation at each of the non-zero generators
                test = range(len(indices))
                for j in range(len(indices)):
                    if j >= nr_of_gens:
                        single_gen_value = 0
                    psspy.machine_chng_2(plant_number, str(j + 1),
                                         [_i, _i, _i, _i, _i, _i],
                                         [single_gen_value, _f, _f, _f, _f, _f, _f,
                                          _f, _f, _f, _f, _f, _f, _f, _f,
                                          _f, _f])

    def reduce_load_by_x_percentage(self, load_reduction_percentage=0.1):  # 0.1 = 10%
        # THIS FUNCTION MUST BE CALLED BEFORE HVDC IMPORT IS ADJUSTED TO APPROPRIATE VALUE, ELSE HVDC OUTAGE SIZE WILL
        # BE REDUCED
        # Reduce all load by x percentage and also the generation in each area. This will hopefully let me decrease
        # inertia of the system

        # Percentage between 0% and 100%
        if load_reduction_percentage > 1 or load_reduction_percentage < 0:
            raise

        bus_numbers_loads = np.array(psspy.aloadint(-1, 1, "NUMBER")[1][0])
        area_numbers_loads = np.array(psspy.aloadint(-1,1, "AREA")[1][0])
        active_powers_loads = np.array(psspy.aloadcplx(-1, 1, "MVAACT")[1][0]).real

        # Reduce loads by given percentage
        bus_number_load = 0
        load_id = 0
        for i in range(len(bus_numbers_loads)):
            if bus_number_load != bus_numbers_loads[i]:  # new load
                load_id = 1
                bus_number_load = bus_numbers_loads[i]
            else:  # Same machine as previous
                load_id += 1
            new_load_setpoint = active_powers_loads[i]*(1-load_reduction_percentage)
            psspy.load_chng_4(bus_number_load, str(load_id), [_i, _i, _i, _i, _i, _i], [new_load_setpoint, _f, _f, _f, _f, _f])

        # Make a list representing how much load has been changed in each area
        area_numbers_unique_sorted = np.unique(area_numbers_loads)
        areas_loads_active_power_reduced = []
        for i in range(len(area_numbers_unique_sorted)):
            indices = np.where(area_numbers_loads == area_numbers_unique_sorted[i])
            areas_loads_active_power_reduced.append(sum(active_powers_loads[indices])*load_reduction_percentage)

        # Reduce generation of generators in each area corresponding to the area load decrease
        for i in range(len(area_numbers_unique_sorted)):
            self.reduce_active_power_gen_by_x_in_area_y(areas_loads_active_power_reduced[i],
                                                        area_numbers_unique_sorted[i])

        # Find the minimum amount of generators needed to supply current generation
        self.reduce_to_minimum_generators_needed_for_generation()

        breakpoint = 1

    def increase_gen_to_decrease_slack_bus_output(self, percentage_increase):
        machine_active_powers = np.array(psspy.amachreal(-1, 1, "PGEN")[1][0])
        machine_bus_numbers = np.array(psspy.amachint(-1, 1, "NUMBER")[1][0])

        current_machine_nr = 0  # machine_chng_2 needs both number and ID
        machine_id = 0  # which complicates the iteration through the machines
        for i in range(len(machine_bus_numbers)):
            if machine_bus_numbers[i] != 3300:  # Not the slack bus
                if current_machine_nr != machine_bus_numbers[i]:  # new machine
                    machine_id = 1
                    current_machine_nr = machine_bus_numbers[i]
                else:  # Same machine as previous
                    machine_id = machine_id + 1

                psspy.machine_chng_2(current_machine_nr, str(machine_id),
                                     [_i, _i, _i, _i, _i, _i],
                                     [machine_active_powers[i]*(1 + percentage_increase), _f, _f, _f, _f, _f, _f,
                                      _f, _f, _f, _f, _f, _f, _f, _f,
                                      _f, _f])

    def reduce_active_power_gen_by_x_in_area_y(self, active_power_reduction_for_area=100, area_number=11):

        bus_numbers = np.array(psspy.abusint(-1, 1, "NUMBER")[1][0])  # Get all bus numbers
        bus_areas = np.array(psspy.abusint(-1, 1, "AREA")[1][0])

        # Find generation capacity of this specific area
        areas_of_plants = np.array(psspy.agenbusint(-1, 1, "AREA")[1][0])
        indices = np.where(areas_of_plants == area_number)[0]
        gen_of_plants = np.array(
            psspy.agenbusreal(-1, 1, "PGEN")[1][0])  # Retrieve active power generation of all plants
        gen_in_hvdc_bus_area = gen_of_plants[indices]  # Only include area of hvdc bus
        gen_capacities_of_plants = np.array(
            psspy.agenbusreal(-1, 1, "PMAX")[1][0])  # Retrieve generation capacity of all plants
        gen_capacity_in_hvdc_bus_area = gen_capacities_of_plants[indices]
        sum_gen_in_hvdc_bus_area = sum(gen_in_hvdc_bus_area)  # Currently active generation in the HVDC area
        sum_gen_of_plants = sum(gen_of_plants)  # Currently active generation in the system

        # psspy.amachint can not return "AREA", thus we need a workaround using machine numbers and bus areas
        # Find "AREA" of each machine
        machine_numbers = np.array(psspy.amachint(-1, 1, "NUMBER")[1][0])
        machine_areas = np.zeros(len(machine_numbers))
        for i in range(len(machine_numbers)):  # For loop to assign area numbers
            machine_areas[i] = bus_areas[np.where(bus_numbers == machine_numbers[i])]

        indices = np.where(machine_areas == area_number)[0]  # Indices of machines in the HVDC bus area
        machine_generation = np.array(psspy.amachreal(-1, 1, "PGEN")[1][0])  # Read PGEN of machines
        machine_generation_capacity = np.array(psspy.amachreal(-1, 1, "PMAX")[1][0])  # Read PMAX of machines
        machine_gen_change = machine_generation  # The generation that can be reduced when import is increased. NB:
        # machine_generation includes ALL machines, not just the ones in HVDC area

        current_machine_nr = 0  # machine_chng_2 needs both number and ID
        machine_id = 0  # which complicates the iteration through the machines

        # Distribute generation change in local region
        if abs(sum_gen_in_hvdc_bus_area) > abs(
                active_power_reduction_for_area):  # Distribute gen change over local
            # area if the generators can take it all
            gen_change_ratio = abs(active_power_reduction_for_area) / sum_gen_in_hvdc_bus_area
            machine_gen_change = machine_gen_change * gen_change_ratio

            # Set each generator to unique value
            for i in indices:  # Iterate through each in-service machine and set generation
                if current_machine_nr != machine_numbers[i]:  # new machine
                    machine_id = 1
                    current_machine_nr = machine_numbers[i]
                else:  # Same machine as previous
                    machine_id = machine_id + 1

                psspy.machine_chng_2(current_machine_nr, str(machine_id),
                                     [_i, _i, _i, _i, _i, _i],
                                     [machine_generation[i] - machine_gen_change[i], _f, _f, _f, _f, _f, _f,
                                      _f, _f, _f, _f, _f, _f, _f, _f,
                                      _f, _f])

        else:  # Distribute over whole system

            # First set all generation to 0 in local area
            for i in indices:  # Iterate through machines in local area and set to zero generation
                if current_machine_nr != machine_numbers[i]:  # new machine
                    machine_id = 1
                    current_machine_nr = machine_numbers[i]
                else:  # Same machine as previous
                    machine_id = machine_id + 1

                psspy.machine_chng_2(current_machine_nr, str(machine_id),
                                     [_i, _i, _i, _i, _i, _i],
                                     [0.0, _f, _f, _f, _f, _f, _f, _f,
                                      _f, _f, _f, _f, _f, _f, _f, _f, _f])

            # After setting to 0 in local area, recalculate slack to distribute over system
            active_power_reduction_for_area = active_power_reduction_for_area + \
                                                 sum(machine_gen_change[indices])  # Recalculate the slack that
            # needs to be distributed
            system_remaining_generation_capacity = sum_gen_of_plants - sum(
                machine_gen_change[indices])  # Recalculate remaining generation capacity of
            # system
            machine_gen_change[indices] = 0  # Generators at 0 = can't take more slack
            gen_change_ratio = abs(active_power_reduction_for_area) / system_remaining_generation_capacity
            machine_gen_change = machine_gen_change * gen_change_ratio

            # Distribute this new slack across the rest of the system
            machine_id = 0
            current_machine_nr = 0
            for i in range(len(machine_gen_change)):  # Increase generation of all generators in system
                if current_machine_nr != machine_numbers[i]:  # New machine
                    machine_id = 1
                    current_machine_nr = machine_numbers[i]
                else:  # Same machine as previous
                    machine_id = machine_id + 1
                psspy.machine_chng_2(current_machine_nr, str(machine_id),
                                     [_i, _i, _i, _i, _i, _i],
                                     [machine_generation[i] - machine_gen_change[i], _f, _f, _f, _f, _f, _f,
                                      _f, _f, _f, _f, _f, _f, _f, _f, _f, _f])
        self.reduce_to_minimum_generators_needed_for_generation()  # All generation has now been distributed

        # Now all slack has been distributed

    def run_static_load_flow(self):
        """
            Runs a Fixed Slope Decoupled Newton-Raphson power flow study. Used before running dynamic simulations, so
            that any model changes are taken into account in the power flows.
        """
        psspy.fdns([0, 0, 0, 1, 1, 1, 99, 0])  # Fixed slope decoupled Newton-Raphson

    # ========================= Functions for plotting and saving ======================================================
    def read_results(self):
        """
            Reads the output-file from the time-domain simulation and stores it within the case-object.
        """
        chnf = dyntools.CHNF(self.output_file + ".out")
        self.sh_ttl, self.ch_id, self.ch_data = chnf.get_data()

    def plot_results_and_save(self):
        """
            Plots the results from the time-domain simulation. Depending on case configuration, the plots are either
            directly presented to the user or only saved for later viewing. In addition, the files may be stored locally
            in the folder which the function is run, or saved in the ShareLatex folder.
        """
        # Directory of plot files
        reset_cwd_for_next_case = os.path.join(os.getcwd(), "PycharmProject")
        os.chdir(r"C:\Users\Espen\Dropbox\Apps\ShareLaTeX\Master Thesis\Plots")
        os.chdir(reset_cwd_for_next_case)
        if self.save_plots_local_folder:  # Local placement, save in folder near script being run
            os.chdir("Plots")  # To save in a separate folder for Plots
        else:  # Save in folder for LaTeX-document to include in thesis
            os.chdir(r"C:\Users\Espen\Dropbox\Apps\ShareLaTeX\Master Thesis\Plots")


    # Plot programmable generation
        if self.monitor_data_manually:
            for i in range(len(self.programmable_generation)):
                self.programmable_generation[i].plot(self.save_plots_local_folder)

    # Plot machine quantities previously specified by the user that he or she wishes to monitor
        machine_quantities = ("ANGLE", "PELEC", "QELEC", "ETERM", "EFD", "PMECH",
                              "SPEED", "XADIFD", "ECOMP", "VOTHSG", "VREF", "VUELL",
                              "VOEL", "GREF", "LCREF", "WVLCTY", "WTRBSP", "WPITCH",
                              "WAEROT", "WROTRV", "WROTRI", "WPCMND", "WQCMND")

        machine_user_input = np.unique(self.machine_monitor_overview[:, 1])
        for i in range(len(machine_user_input)):
            plot_title = machine_quantities[int(machine_user_input[i]) - 1]
            plt.figure(plot_title)
            indices = np.where(self.machine_monitor_overview[:, 1] == machine_user_input[i])[0]  # Could maybe be simplified by using machine_monitor_overview column 0?
            if plot_title == "SPEED":  # To not plot all the SPEED channels used for COI-frequency calculation
                new_indices = []
                for j in range(len(indices)):
                    if any(self.machine_monitor_overview[indices[j]][2] == self.machines_to_monitor):
                        new_indices.append(indices[j])
                indices = new_indices
            for j in range(len(indices)):
                if plot_title == "SPEED":
                    for k in range(len(self.ch_data[indices[j]+1])):  # Adjust data so that frequency is appropriate
                        self.ch_data[indices[j]+1][k] = 50 + self.ch_data[indices[j]+1][k]*50
                plt.plot(self.ch_data['time'], self.ch_data[indices[j]+1])

            if self.calculated_coi_frequency and plot_title == "SPEED":
                plt.plot(self.ch_data['time'], self.coi_frequency, 'r--')
                # Maybe problem generating legends?

            plt.xlabel("Time (s)")
            plt.ylabel(self._generate_ylabel(plot_title))
            if self.calculated_coi_frequency and plot_title == "SPEED":
                test = self._generate_machine_legends(indices)
                test.append("COI-frequency")
                plt.legend(test)
            else:
                plt.legend(self._generate_machine_legends(indices))
            plt.grid()
            plt.savefig(machine_quantities[int(machine_user_input[i]-1)] + "_" + self.filename)

    # Plot bus quantities previously specified by the user that he or she wishes to monitor
        bus_quantities = ("BSFREQ", "LOADP", "LOADQ", "BSVOLT")

        buses_user_input = np.unique(self.bus_monitor_overview[:, 1])
        for i in range(len(buses_user_input)):
            plot_title = bus_quantities[int(buses_user_input[i]) - 1]
            plt.figure(plot_title)
            indices = np.where(self.bus_monitor_overview[:, 1] == buses_user_input[i])[0]
            if plot_title == "BSFREQ":  # To not plot all the BSFREQ channels used for PGU calculations
                new_indices = []
                for j in range(len(indices)):
                    if any(self.bus_monitor_overview[indices[j]][2] == self.buses_to_monitor):
                        new_indices.append(indices[j])
                indices = new_indices
            for j in range(len(indices)):
                quantity_channel_index = self.bus_monitor_overview[indices[j]][0]
                if plot_title == "BSFREQ":
                    for k in range(len(self.ch_data[quantity_channel_index])):  # Adjust from p.u. to Hz
                        self.ch_data[quantity_channel_index][k] *= 50
                        self.ch_data[quantity_channel_index][k] += 50
                plt.plot(self.ch_data['time'], self.ch_data[quantity_channel_index])

            plt.xlabel("Time (s)")
            plt.ylabel(self._generate_ylabel(plot_title))
            plt.legend(self._generate_bus_legends(indices))
            plt.grid()
            plt.savefig(bus_quantities[int(buses_user_input[i] - 1)] + "_" + self.filename)

    # Plot branch flows
        branch_legends = []
        for i in range(len(self.branch_p_channels)):
            plt.figure("BRANCHFLOWS")
            plt.plot(self.ch_data["time"], self.ch_data[self.branch_p_channels[i][0]])
            branch_legends.append("From "+str(self.branch_p_channels[i][1])+" to "+str(self.branch_p_channels[i][2]))
        plt.xlabel("Time (s)")
        plt.ylabel("Branch active power flows [MW]")
        plt.legend(branch_legends)
        plt.grid()

    # Show plots to user or save as .png directly
        if self.present_plots:
            plt.show()
        os.chdir(reset_cwd_for_next_case)  # Sets directory back to original, so that subsequent case runs do not
        # disturb each other

    def save_network_data(self):
        """
            Saves the network data so the model alterations can be tested in PSSE along with PssPy.
        """
        self.run_static_load_flow()
        psspy.save(os.path.join(os.getcwd(), self.filename+".sav"))  # Static updates
        psspy.dyda(0, 1, [2, 1, 0], 0, os.path.join(os.getcwd(), self.filename +".dyr"))  # Dynamic updates

    # ========================= Auxiliary functions ====================================================================
    def calculate_coi_frequency(self, time_1=1.0, time_2=2.0):
        """
            Calculates the Centre of Inertia-frequency and returns the ROCOF of the system between specified times
            Input:
                - time_1: start of time interval in seconds [s] over which ROCOF is calculated
                - time_2: end of time interval in seconds [s] over which ROCOF is calculated
            Output:
                - min_freq: The lowest minimum COI-frequency encountered during the simulation
            To-do:
                - Add plot of ROCOF-value so it can be displayed in the COI-frequency plot?
                - Calculate system inertia and return it as value? (Needs to know size of outage, can be implemented
                by values saved in object

        """

        self.calculated_coi_frequency = True  # Set to true so plot function can plot frequencies relative to COI frequency

        # First, find all inertia constants H (for using to calculate COI-frequency)
        all_online_machines = np.array(psspy.amachint(-1, 1, "NUMBER")[1][0])
        H_list = []  # Acquire all inertia constants for calculation of COI-frequency
        machine_nr = 0
        machine_id = 0
        for i in range(len(all_online_machines)):
            if machine_nr != all_online_machines[i]:  # new machine
                machine_id = 1
                machine_nr = all_online_machines[i]
            else:  # Same machine as previous
                machine_id += 1
            indx = psspy.mdlind(machine_nr, str(machine_id), "GEN", "CON")[1]
            gen_name = psspy.mdlnam(machine_nr, str(machine_id), "GEN")[1]
            if gen_name[0:6] == "GENSAL":
                indx += 3  # Offset to get the H constant
            elif gen_name[0:6] == "GENROU":
                indx += 4  # Offset to get the H constant
            else:
                raise
            H = psspy.dsrval("CON", indx)[1]
            H_list.append(H)  #

    # H for all online machines has been found. Now use to find COI-frequency
        coi_frequency = np.zeros((len(self.ch_data['time']), 1))  # Initialize matrix for COI-frequency with zeros
        for i in range(len(all_online_machines)):  # For each monitored machine, use H weighting to calculate COI-frequency
            machine_nr = all_online_machines[i]

            machine_nr_match = np.where(self.machine_monitor_overview[:, 2] == machine_nr)[0]
            machine_frequency_monitoring_match = np.where(self.machine_monitor_overview[:, 1] == 7)[0]
            for j in range(len(machine_nr_match)):
                # THIS SEARCHING MAY CAUSE ISSUES; TAKE NOTE IF PROBLEMS OCCUR
                temp_index = np.where(machine_frequency_monitoring_match == machine_nr_match[j])[0]
                if len(temp_index) > 0:
                    freq_chan_index = self.machine_monitor_overview[machine_frequency_monitoring_match[temp_index[0]]][0]
                    break

            for j in range(len(self.ch_data['time'])):
                coi_frequency[j] += (50 + self.ch_data[int(freq_chan_index)][j] * 50) * (H_list[i] / sum(H_list))
        self.coi_frequency = coi_frequency

        #Find initial ROCOF
        index_1 = np.searchsorted(self.ch_data['time'], time_1) + 1
        index_2 = np.searchsorted(self.ch_data['time'], time_2) + 1
        rocof = (coi_frequency[index_2] - coi_frequency[index_1]) / (time_2 - time_1)  # ROCOF, [Hz/s]
        print ("The initial ROCOF is: " + str(rocof) + "[Hz/s]")


        if False:
            plot_title = "COI-frequency"
            plt.figure(plot_title)
            plt.plot(self.ch_data["time"], coi_frequency)
            plt.xlabel("Time (s)")
            plt.ylabel("COI frequency Hz")
            plt.grid()
            plt.show()



        # Assume that the first time instance in which an outage occurs is the only one we care about, and use the size
        # of the outage there to calculate system inertia based upon the ROCOF response
        contingency_time = np.sort(np.array(self.contingency_time))
        contingency_size = np.array(self.contingency_size)
        total_initial_outage_size = 0
        for i in range(len(contingency_time)):
            print("contingency_time[i] is: "+str(int(contingency_time[i])))
            if int(contingency_time[i]) == int(contingency_time[0]):  # Assuming that happening in the same second are
                # simultaneous
                total_initial_outage_size += contingency_size[i]

        # Calculate total inertia in MWs from inertia constants
        machines_mbase = np.array(psspy.amachreal(-1, 1, "MBASE")[1][0])
        inertia_list = []
        for i in range(len(all_online_machines)):
            inertia_list.append(H_list[i] * machines_mbase[i])
        total_calculated_inertia = sum(inertia_list)

        # Calculate total inertia in MWs from ROCOF
        total_estimated_inertia = total_initial_outage_size * 50 / (2*abs(rocof[0]))

        # Now the size of the initial outage has been found and can be used to estimate inertia by using the ROCOF
        print("Outage occurs at time:                                       "+str(contingency_time[0])+" [s]")
        print("Total size of outage:                                        "+str(total_initial_outage_size)+" [MW]")
        print("The total inertia by using inertia constants is:             " + str(total_calculated_inertia)+" [MWs]")
        print("The total inertia estimated using ROCOF approximation is:    " + str(total_estimated_inertia)+" [MWs]")
        #print("The estimated frequency bias is: "+str(frequency_bias)+" [MW/Hz]")

        freq_ss = self.coi_frequency[len(self.coi_frequency)-1]
        freq_nadir = min(self.coi_frequency)
        return [freq_nadir, freq_ss]

    def disconnect_inactive_generators(self):
        """
            Disconnects generators that are generating below a certain % (currently 5%) of their max capacity. Useful
            for cases where import is high and some generators are set to 0 generation due to the redistribution
            algorithm when setting HVDC link power.

            A potential issue: redistribution does not take into account generators that are off-line, and is thus not
            taking into account all generators in the system. Fix?
                """

        percentage_limit = 0.05  # Percentage limit, lower is disconnected
        gen = np.array(psspy.agenbusreal(-1, 1, "PGEN")[1][0])  # Retrieve generation of generators
        gen_max = np.array(psspy.agenbusreal(-1, 1, "PMAX")[1][0])  # Retrieve max generation of generators
        gen_percentage = gen / gen_max
        indices = np.where(gen_percentage < 0.05)[0]  # Indices of plants greater than 5% is disconnected

        plant_numbers = np.array(psspy.agenbusint(-1, 1, "NUMBER")[1][0])
        machine_numbers = np.array(psspy.amachint(-1, 1, "NUMBER")[1][0])  # Bus numbers from machine data
        for i in range(len(indices)):  # Loop to disconnect all machines with too low generation
            bus_nr = plant_numbers[indices[i]]  # Bus number from plant data
            bus_ids = sum(machine_numbers == bus_nr)
            for j in range(bus_ids):
                # Disconnect in "Network Data"
                bus_id = str(j + 1)  # Or r"""1""" ?
                psspy.machine_chng_2(bus_nr, bus_id, [0, _i, _i, _i, _i, _i],
                                     [_f, _f, _f, _f, _f, _f, _f, _f, _f, _f, _f, _f, _f, _f, _f, _f, _f])

                # Disconnect in "Dynamics data"?
                # (In retrospect not sure if following lines of code are necessary)
                psspy.plmod_status(bus_nr, bus_id, 1, 0)  # Generator model disconnected
                # psspy.plmod_status(bus_nr, bus_id, 3, 0)  # Stabilizer model disconnected
                psspy.plmod_status(bus_nr, bus_id, 6, 0)  # Excitation system model disconnected
                psspy.plmod_status(bus_nr, bus_id, 7, 0)  # Turbine-governor model disconnected

    def disconnect_inactive_generators_fixed(self, percentage_limit = 0.05):
        """
            Disconnects generators that are generating below a certain % (currently 5%) of their max capacity. Useful
            for cases where import is high and some generators are set to 0 generation due to the redistribution
            algorithm when setting HVDC link power.

            A potential issue: redistribution does not take into account generators that are off-line, and
            is thus not taking into account all generators in the system. Fix?
            Input:
                percentage_limit: if output of a generator is lower than a certain percentage (percentage limit) of its
                maximum generation it is disconnected.
        """

        gen = np.array(psspy.amachreal(-1, 1, "PGEN")[1][0])  # Retrieve generation of generators
        gen_max = np.array(psspy.amachreal(-1,1,"PMAX")[1][0])  # Retrieve max generation of generators
        gen_percentage = gen / gen_max
        machine_bus_numbers = np.array(psspy.amachint(-1, 1, "NUMBER")[1][0])  # Bus numbers from machine data

        machine_bus_number = 0
        machine_id = 0
        for i in range(len(machine_bus_numbers)):  # Loop to disconnect all machines with too low generation
            if machine_bus_number != machine_bus_numbers[i]:  # new machine
                machine_id = 1
                machine_bus_number = machine_bus_numbers[i]
            else:  # Same machine as previous
                machine_id += 1

            # See if generator is at a very low output; if yes, disconnect it.
            gen_percentage = gen[i] / gen_max[i]
            if gen_percentage < percentage_limit:
                # Disconnect in "Network Data"
                psspy.machine_chng_2(machine_bus_number, str(machine_id), [0, _i, _i, _i, _i, _i],
                                     [_f, _f, _f, _f, _f, _f, _f, _f, _f, _f, _f, _f, _f, _f, _f, _f, _f])

                # Disconnect in "Dynamics data"? Not sure if following lines are necessary:
                psspy.plmod_status(machine_bus_number, str(machine_id), 1, 0)  # Generator model disconnected
                # psspy.plmod_status(bus_nr, str(machine_id), 3, 0)  # Stabilizer model disconnected
                psspy.plmod_status(machine_bus_number, str(machine_id), 6, 0)  # Excitation system model disconnected
                psspy.plmod_status(machine_bus_number, str(machine_id), 7, 0)  # Turbine-governor model disconnected


    def return_frequency_bias(self):
        """
            Returns the effective frequency bias of the generators in the system for the case study by comparing
            generator active power output pre-fault and steady-state post-fault.
        """
        hygov_buses = [3115, 3115, 3115, 3245, 3249, 3249, 3249, 3249, 3249, 3249, 3249, 5100, 5300, 5300, 5400, 5400,
                       5500, 5600, 5600, 6000, 6100, 6100, 6100, 6100, 6100, 6500, 6500, 6500, 6500, 6700, 6700, 7100,
                       7100, 7100]  # Only hydro generators participate in primary response

        machine_bus_number = 0
        machine_id = 0
        for i in range(len(hygov_buses)):
            if machine_bus_number != hygov_buses[i]:  # new machine
                machine_id = 1
                machine_bus_number = hygov_buses[i]
            else:  # Same machine as previous
                machine_id += 1


        ss_frequency = self.coi_frequency[len(self.coi_frequency)-1]
        frequency_difference = 50 - ss_frequency
        frequency_bias = power_difference / frequency_difference
        return frequency_bias

# ========================= Private functions =========================

    def _find_unique_times(self):
        """
            Finds all the times (in seconds) at which faults have been set by the user to occur in the dynamic
            simulation/case study. Sorts the list and stores it in the case-object. Used to have a sequence of which
            time to run/stop simulation.
        """
        time = []
        for i in range(len(self.events_overview)):
            time.append(self.events_overview[i][0])
        time.sort()
        return np.array(time)

    def _generate_ylabel(self, quantity="ANGLE"):
        """
            Used in the presentation of results to generate the label for the y-axis. This is for the machine-related
            quantities. They are substituted from PSSE-descriptions to proper labels.
        """
        return {  # Define dictionary
            # Machine quantities
            "ANGLE": "Angle (deg)",
            "PELEC": "Active power (p.u.)",
            "QELEC": "Reactive power (p.u.)",
            "ETERM": "Terminal voltage (p.u.)",
            "EFD": "UNDEFINED YLABEL",
            "PMECH": "Mechanical power (p.u.)",
            "SPEED": "Frequency [Hz]",
            "XADIFD": "UNDEFINED YLABEL",
            "ECOMP": "UNDEFINED YLABEL",
            "VOTHSG": "UNDEFINED YLABEL",
            "VREF": "Reference voltage (p.u)",
            "VUELL": "UNDEFINED YLABEL",
            "VOEL": "UNDEFINED YLABEL",
            "GREF": "UNDEFINED YLABEL",
            "LCREF": "UNDEFINED YLABEL",
            "WVLCTY": "UNDEFINED YLABEL",
            "WTRBSP": "UNDEFINED YLABEL",
            "WPITCH": "UNDEFINED YLABEL",
            "WAEROT": "UNDEFINED YLABEL",
            "WROTRV": "UNDEFINED YLABEL",
            "WROTRI": "UNDEFINED YLABEL",
            "WPCMND": "UNDEFINED YLABEL",
            "WQCMND": "UNDEFINED YLABEL",

            # Bus quantities
            "BSFREQ": "Bus frequency [Hz]",
            "LOADP": "Bus load active power (p.u.)",
            "LOADQ": "Bus load reactive power (p.u.)",
            "BSVOLT": "Bus voltage (p.u.)"

        }[quantity]  # Return key quantity

    def _generate_machine_legends(self, indices):
        """
            Generates legends for the plots of machine channel quantities.
        """
        legend = []
        for i in range(len(indices)):
            legend.append("Bus " + str(int(self.machine_monitor_overview[indices[i]][2])))

        return legend

    def _generate_bus_legends(self, indices):
        """
            Generates legends for the plots of machine channel quantities.
        """
        legend = []
        for i in range(len(indices)):
            legend.append("Bus " + str(int(self.bus_monitor_overview[indices[i]][2])))

        return legend

# Unknown purpose functions:

    # ==================== Discarded functions below here ====================

    def redistribute_slack(self, slack_bus_number=3300):
        """
            Write this documentation
        """

        # Use same methodology as described in setHvdcToMax
        # Continue until the excess slack is less than 1% of max slack generator capacity

        mach_gen = np.array(psspy.amachreal(-1, 1, "PGEN")[1][0])  # Read PGEN of machines
        mach_gen_cap = np.array(psspy.amachreal(-1, 1, "PMAX")[1][0])  # Read PMAX of machines
        mach_slack = mach_gen_cap - mach_gen  # NB: mach_slack includes ALL machines, not just the ones in HVDC area
        mach_indices = np.array(psspy.amachint(-1, 1, "NUMBER")[1][0])
        slack_mach_indices = np.where(mach_indices == slack_bus_number)
        slack_mach_exceeded = sum(mach_slack[slack_mach_indices])
        machine_numbers = np.array(psspy.amachint(-1, 1, "NUMBER")[1][0])

        # Distribute this new slack across the rest of the system
        machine_id = 0
        current_machine_nr = 0
        for i in range(len(mach_slack)):  # Increase generation of all generators in system
            if current_machine_nr != machine_numbers[i]:  # New machine
                machine_id = 1
                current_machine_nr = machine_numbers[i]
            else:  # Same machine as previous
                machine_id = machine_id + 1
            psspy.machine_chng_2(current_machine_nr, str(machine_id),
                                 [_i, _i, _i, _i, _i, _i],
                                 [mach_gen[i] + mach_slack[i], _f, _f, _f, _f, _f, _f,
                                  _f, _f, _f, _f, _f, _f, _f, _f,
                                  _f,
                                  _f])

            # All generation has now been distributed
            # Run new power flow and see if limits are exceeded more than 1% of slack bus
            # Recursive algorithm if
            self.run_static_load_flow()
            pass

        pass


